/* Letris = Linux Tetris  :) */
#include <time.h>
#include <vga.h>
#include <stdio.h>
#include <termios.h>
#include <curses.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/time.h>
#include <assert.h>

/* Algunos colores y valores validos para G640x480x256 */
#define WHITE 15 
#define RED 4
#define BLUE 1
#define GREEN 2
#define YELLOW 14
#define BLACK 0
#define CYAN 3
#define LIGHTCYAN 11
#define BROWN 6
#define LIGHTBLUE 9 
#define LIGHTGREEN 10
#define MAGENTA 5
#define LIGHTGRAY 7
#define GRAY 8

float level; // Level
int score=0; // Puntaje alcanzado
int lado=15; // Lado del cuadrado. Valor entero entre 1 y 15 pixels
int color_borde=WHITE;
int m[12][34];
int l[5][6]; // Matriz que contiene a la letra
int f[4][4];  // Matriz que contiene a la ficha
int color;  // Color de la ficha

static struct termios initial_settings, new_settings;
static int peek_character = -1;

void init_keyboard()
{
  tcgetattr(0,&initial_settings);
  new_settings = initial_settings;
  new_settings.c_lflag &= ~ICANON;
  new_settings.c_lflag &= ~ECHO;
  new_settings.c_lflag &= ~ISIG;
  new_settings.c_cc[VMIN] = 1;
  new_settings.c_cc[VTIME] = 0;
  tcsetattr(0, TCSANOW, &new_settings);
}

void close_keyboard()
{
  tcsetattr(0, TCSANOW, &initial_settings);
}

int kbhit()
{
  char ch;
  int nread;
  if(peek_character != -1) return 1;
  new_settings.c_cc[VMIN]=0;
  tcsetattr(0, TCSANOW, &new_settings);
  nread = read(0, &ch,1);
  new_settings.c_cc[VMIN]=1;
  tcsetattr(0, TCSANOW, &new_settings);
  if(nread==1)
  {
    peek_character=ch;
    return 1;
  }
  return 0;
}

int readch()
{
  char ch;
  if(peek_character!=-1)
  {
    ch=peek_character;
    peek_character=-1;
    return ch;
  }
  read(0, &ch, 1);
  return ch;
}

void msleep(unsigned long msecs)
{
	struct timeval timeout;
	int oops;
	timeout.tv_sec = 0;
	timeout.tv_usec = msecs;
       	// timeout.tv_sec = msecs / 1000;
	// timeout.tv_usec = msecs % 1000;
       	oops = select(0, (fd_set *)0, (fd_set *)0, (fd_set *)0, &timeout);
	assert( oops >= 0 );
}

void rect(int x1, int y1, int x2, int y2)
{
  vga_drawline(x1,y1,x1,y2);
  vga_drawline(x1,y2,x2,y2);
  vga_drawline(x2,y2,x2,y1);
  vga_drawline(x2,y1,x1,y1);
}

void rectlleno(int x1, int y1, int x2, int y2)
{
   int y;
   for(y=y1;y<=y2;y++)
     vga_drawline(x1,y,x2,y);
}

void cuadro(int x,int y,int color1,int color2)
{
  int a,b,izq,abj;
  izq=10+(15-lado)*10;  // Margen izquierdo segun el tamano
  abj=465-(15-lado)*31; // Margen inferior segun el tamano
  a=izq+lado*x;
  b=abj-((32-y)*lado);
  vga_setcolor(color1);
  rectlleno(a,b,a+lado-1,b+lado-1); 
  vga_setcolor(color2);
  rect(a,b,a+lado-1,b+lado-1);
}

borrar(int a, int b) // Borra si la matriz M == 0 en la posicion
{
  int x,y;
  for (x=0;x<=3;x++)
	for (y=0;y<=3;y++)
	  if (m[x+a][y+b]==0) cuadro(x+a,y+b,BLACK,BLACK);
  return 0;
}

void letra(int posx, int posy, int n, int tam, int cint, int cout)
{
  int x,y;
  for(x=0;x<=4;x++)
    for(y=0;y<=5;y++) l[x][y]=0; // Inicializa la letra
  switch (n)
    {
      case 'T': { for(y=0;y<=5;y++) l[2][y]=1; for(x=0;x<=4;x++) l[x][0]=1; break; }
      case 'E': { for(y=0;y<=5;y++) l[0][y]=1; for(x=1;x<=4;x++) l[x][0]=1; for(x=1;x<=4;x++) l[x][2]=1; for(x=1;x<=4;x++) l[x][5]=1; break; }
      case 'R': { for(x=0;x<=3;x++) l[x][0]=1; for(y=1;y<=5;y++) l[0][y]=1; for(x=0;x<=3;x++) l[x][2]=l[3][1]=l[1][3]=l[2][4]=l[3][5]=1; break; }
      case 'I': { for(y=0;y<=5;y++) l[2][y]=1; break; }
      case 'S': { for(x=0;x<=4;x++) l[x][0]=1; for(x=0;x<=4;x++) l[x][2]=1; for(x=0;x<=4;x++) l[x][5]=1; l[0][1]=l[4][3]=l[4][4]=1; break; }
      case 'P': { for(y=0;y<=5;y++) l[0][y]=1; for(x=1;x<=4;x++) l[x][0]=1; for(x=1;x<=4;x++) l[x][2]=1; l[4][1]=1; break; }
      case 'L': { for(y=0;y<=5;y++) l[0][y]=1; for(x=1;x<=4;x++) l[x][5]=1; break; }
      case 'A': { for(x=0;x<=3;x++) l[x][3]=1; for(y=2;y<=5;y++) { l[0][y]=l[4][y]=1; } l[2][0]=l[1][1]=l[3][1]=1; break; }
      case 'Y': { for(y=2;y<=5;y++) l[2][y]=1; l[0][0]=l[1][1]=l[4][0]=l[3][1]=1; break; }
      case 'X': { l[0][0]=l[1][1]=l[2][2]=l[4][0]=l[3][1]=l[1][4]=l[0][5]=l[2][3]=l[4][5]=l[3][4]=1; break; }
      case 'N': { for(y=0;y<=5;y++) l[0][y]=l[4][y]=1; l[1][1]=l[2][2]=l[2][3]=l[3][4]=1; break; }
      case 'F': { for(y=0;y<=5;y++) l[0][y]=1; for(x=1;x<=4;x++) l[x][0]=1; for(x=1;x<=4;x++) l[x][2]=1; break; }
      case 'O': { for(y=1;y<=4;y++) l[0][y]=l[4][y]=1; for(x=1;x<=3;x++) l[x][0]=l[x][5]=1; break; }
      case 'U': { for(y=0;y<=5;y++) l[0][y]=l[4][y]=1; for(x=1;x<=4;x++) l[x][5]=1; break; }
      case 'C': { for(y=1;y<=4;y++) l[0][y]=1; for(x=1;x<=3;x++) l[x][0]=l[x][5]=1; break; }
      case 'D': { for(y=0;y<=5;y++) l[0][y]=1; for(x=1;x<=2;x++) l[x][0]=l[x][5]=1; l[3][1]=l[4][2]=l[4][3]=l[3][4]=1; break; }
      case '-': { for(x=0;x<=5;x++) l[x][2]=1; break; }
      case 'G': { for(y=1;y<=4;y++) l[0][y]=1; for(x=1;x<=3;x++) l[x][0]=l[x][5]=1; l[4][4]=l[4][3]=l[3][3]=1; break; }
      case 'H': { for(y=0;y<=5;y++) l[0][y]=l[4][y]=1; for(x=1;x<=3;x++) l[x][2]=1; break; }
      case 'W': { for(y=0;y<=5;y++) l[0][y]=l[4][y]=1; l[1][4]=l[2][3]=l[3][4]=1; break; }
      case '4': { for(y=0;y<=5;y++) l[4][y]=1; for(x=0;x<=3;x++) l[x][2]=1; l[0][0]=l[0][1]=1; break; }
      case '6': { for(y=0;y<=5;y++) l[0][y]=1; for(x=1;x<=4;x++) l[x][2]=l[x][5]=l[x][0]=1; l[4][4]=l[4][3]=1; break; }
      case '8': { for(y=0;y<=5;y++) l[0][y]=l[4][y]=1; for(x=1;x<=3;x++) l[x][0]=l[x][5]=l[x][2]=1; break; }
      case '5': { for(x=0;x<=4;x++) l[x][0]=1; for(x=0;x<=4;x++) l[x][2]=1; for(x=0;x<=4;x++) l[x][5]=1; l[0][1]=l[4][3]=l[4][4]=l[0][4]=1; break; }
      case '2': { for(x=0;x<=4;x++) l[x][0]=l[x][5]=l[x][2]=1; l[4][1]=l[0][4]=l[0][3]=1; break; }
      case 'M': { for(y=0;y<=5;y++) l[0][y]=l[4][y]=1; l[1][1]=l[2][2]=l[3][1]=1; break; }
      case 'V': { for(y=0;y<=3;y++) l[0][y]=l[4][y]=1; l[1][4]=l[2][5]=l[3][4]=1; break; }
      case '1': { for(y=0;y<=5;y++) l[2][y]=1; l[1][5]=l[3][5]=l[1][0]=1; break; }
      case '3': { for(y=0;y<=5;y++) l[4][y]=1; for(x=0;x<=3;x++) l[x][0]=1; for(x=0;x<=3;x++) l[x][2]=1; for(x=0;x<=3;x++) l[x][5]=1; break; }
      case '7': { for(y=0;y<=5;y++) l[4][y]=1; for(x=0;x<=3;x++) l[x][0]=1; l[0][1]=1; break; }
      case '9': { for(y=0;y<=5;y++) l[4][y]=1; for(x=0;x<=3;x++) l[x][2]=l[x][5]=l[x][0]=1; l[0][1]=l[0][2]=1; break; }
      case '0': { for(y=0;y<=5;y++) l[0][y]=l[4][y]=1; for(x=1;x<=3;x++) l[x][0]=l[x][5]=1; break; }
    }
  lado=tam;
  for(x=0;x<=4;x++)
    for(y=0;y<=5;y++)
      if (l[x][y]==1) cuadro(posx+x,posy+y,cint,cout);
  lado=15;
}

cerar(int a, int b)  // Cera la posicion de la ficha
{
  int x,y;
  for (x=0;x<=3;x++)
	for (y=0;y<=3;y++)
	  if (f[x][y]!=0) m[x+a][y+b]=0;
  return 0;
}  

dibujar(int a, int b)
{
  int x,y;
  for (x=0;x<=3;x++)
	for (y=0;y<=3;y++)
	  if (f[x][y]==1)
		{
		   cuadro(x+a,y+b,color,color_borde);
		}
  return 0;
}

fijar(int a, int b)
{
  int x,y;
  for (x=0;x<=3;x++)
	for (y=0;y<=3;y++)
	  if (f[x][y]==1)
		{
		  m[x+a][y+b]=color;
		}
  return 0;
}

crear_ficha(int n)  // Constructor
{
  int x,y;
  for(x=0;x<=3;x++)
	for(y=0;y<=3;y++) f[x][y]=0;   // Inicializa la ficha
  switch (n)
	{
	  case 1: f[1][1]=f[1][2]=f[2][1]=f[2][2]=1; color=YELLOW; break;//ficha 1
	  case 2: f[1][2]=f[2][2]=f[2][1]=f[3][1]=1; color=BLUE; break;//ficha 2
	  case 3: f[0][1]=f[1][1]=f[1][2]=f[2][2]=1; color=GREEN; break;//ficha 3
	  case 4: f[1][1]=f[0][2]=f[1][2]=f[2][2]=1; color=CYAN; break;//ficha 4
	  case 5: f[1][0]=f[2][0]=f[2][1]=f[2][2]=1; color=LIGHTBLUE; break;//ficha 5
	  case 6: f[2][1]=f[2][2]=f[2][3]=f[1][3]=1; color=MAGENTA; break;//ficha 6
	  case 7: f[1][0]=f[1][1]=f[1][2]=f[1][3]=1; color=RED; break;//ficha 7
	}
}

rotar_izq()
{
  int x,y,c[4][4];
  for(y=0;y<=3;y++)
	for(x=0;x<=3;x++) c[y][3-x]=f[x][y]; // Rota la ficha y la pone en c
  for(y=0;y<=3;y++)
	for(x=0;x<=3;x++) f[x][y]=c[x][y];  // Pone la ficha rotada en f
  return 0;
}

rotar_der()
{
  int x,y,c[4][4];
  for(y=0;y<=3;y++)
	for(x=0;x<=3;x++) c[3-y][x]=f[x][y]; // Rota la ficha y la pone en c
  for(y=0;y<=3;y++)
	for(x=0;x<=3;x++) f[x][y]=c[x][y]; // Pone la ficha rotada en f
  return 0;
}

void titulo(void)
{
  letra(14,3,'L',12,BLUE,YELLOW);
  letra(20,3,'E',12,BLUE,YELLOW);
  letra(26,3,'T',12,BLUE,YELLOW);
  letra(32,3,'R',12,BLUE,YELLOW);
  letra(36,3,'I',12,BLUE,YELLOW);
  letra(41,3,'S',12,BLUE,YELLOW);
  letra(35,44,'S',4,LIGHTGRAY,GRAY);
  letra(41,44,'T',4,BLUE,YELLOW);
  letra(47,44,'A',4,BLUE,YELLOW);
  letra(53,44,'R',4,BLUE,YELLOW);
  letra(58,44,'T',4,BLUE,YELLOW);
  letra(35,52,'E',4,BLUE,YELLOW);
  letra(41,52,'X',4,LIGHTGRAY,GRAY);
  letra(45,52,'I',4,BLUE,YELLOW);
  letra(49,52,'T',4,BLUE,YELLOW);
  letra(35,60,'I',4,LIGHTGRAY,GRAY);
  letra(40,60,'N',4,BLUE,YELLOW);
  letra(46,60,'F',4,BLUE,YELLOW);
  letra(52,60,'O',4,BLUE,YELLOW);
  letra(35,68,'P',4,LIGHTGRAY,GRAY);
  letra(41,68,'A',4,BLUE,YELLOW);
  letra(47,68,'U',4,BLUE,YELLOW);
  letra(53,68,'S',4,BLUE,YELLOW);
  letra(59,68,'E',4,BLUE,YELLOW);
  letra(35,76,'E',4,LIGHTGRAY,GRAY);
  letra(41,76,'N',4,BLUE,YELLOW);
  letra(47,76,'D',4,BLUE,YELLOW);
  letra(22,150,'S',3,GREEN,YELLOW);
  letra(28,150,'C',3,GREEN,YELLOW);
  letra(33,150,'O',3,GREEN,YELLOW);
  letra(39,150,'R',3,GREEN,YELLOW);
  letra(44,150,'E',3,GREEN,YELLOW);
}

void marco(void)
{
  int x,y,cint=BLACK,cout=BLACK;
  for(y=1;y<=31;y++)
   {
	  cuadro(0,y,LIGHTGRAY,WHITE);  // Lado izquierdo
	  m[0][y]=1;
	  cuadro(11,y,LIGHTGRAY,WHITE); // Lado derecho
	  m[11][y]=1;
   }
  for(x=0;x<=11;x++)
   {
	 cuadro(x,31,LIGHTGRAY,WHITE);  // Base
	 m[x][31]=1; m[x][32]=1; m[x][33]=1;
   }
  for(y=1;y<=30;y++)
	for(x=1;x<=10;x++) cuadro(x,y,BLACK,BLACK); // Borrar el interior
  titulo();
}

void borrarsc(void)  // Borra zona del score
{
  vga_setcolor(BLACK);
  rectlleno(294,446,382,464);
}

void pscore(void)
{
  int score1=score,i=0;
  char s[5];
  borrarsc();
  while (i<=4)
   {
     s[i++]=(score1 % 10)+'0';
     score1=(score1 - (score1 % 10) ) / 10;
   }
  for(i=0;i<=4;i++) letra(55+i*6,150,s[4-i],3,GREEN,YELLOW);
}

void borrargo(void)
{
  vga_setcolor(BLACK);
  rectlleno(400,310,500,390);
}

void pgameover(void)
{
  int i=RED, o=WHITE;
  letra(70,80,'G',4,i,o);
  letra(76,80,'A',4,i,o);
  letra(82,80,'M',4,i,o);
  letra(88,80,'E',4,i,o);
  letra(70,88,'O',4,i,o);
  letra(76,88,'V',4,i,o);
  letra(82,88,'E',4,i,o);
  letra(88,88,'R',4,i,o);
}

void borrarc(void)
{
  vga_setcolor(BLACK);
  rectlleno(250,170,500,400);
}

void pausa(void)
{
  int i=LIGHTGRAY, o=GRAY;
  borrarc();
  letra(35,68,'P',4,i,o);
  letra(41,68,'A',4,i,o);
  letra(47,68,'U',4,i,o);
  letra(53,68,'S',4,i,o);
  letra(59,68,'E',4,i,o);
  vga_getch();
  titulo();
}

void informacion(void)
{
  borrarc();
  letra(35,44,'4',4,LIGHTGRAY,GRAY);
  letra(41,44,'-',4,BLUE,YELLOW);
  letra(47,44,'L',4,BLUE,YELLOW);
  letra(53,44,'E',4,BLUE,YELLOW);
  letra(59,44,'F',4,BLUE,YELLOW);
  letra(65,44,'T',4,BLUE,YELLOW);
  letra(35,52,'6',4,LIGHTGRAY,GRAY);
  letra(41,52,'-',4,BLUE,YELLOW);
  letra(47,52,'R',4,BLUE,YELLOW);
  letra(51,52,'I',4,BLUE,YELLOW);
  letra(55,52,'G',4,BLUE,YELLOW);
  letra(61,52,'H',4,BLUE,YELLOW);
  letra(67,52,'T',4,BLUE,YELLOW);
  letra(35,60,'5',4,LIGHTGRAY,GRAY);
  letra(41,60,'-',4,BLUE,YELLOW);
  letra(47,60,'R',4,BLUE,YELLOW);
  letra(52,60,'O',4,BLUE,YELLOW);
  letra(58,60,'T',4,BLUE,YELLOW);
  letra(64,60,'A',4,BLUE,YELLOW);
  letra(70,60,'T',4,BLUE,YELLOW);
  letra(76,60,'E',4,BLUE,YELLOW);
  letra(35,68,'2',4,LIGHTGRAY,GRAY);
  letra(41,68,'-',4,BLUE,YELLOW);
  letra(47,68,'D',4,BLUE,YELLOW);
  letra(53,68,'O',4,BLUE,YELLOW);
  letra(59,68,'W',4,BLUE,YELLOW);
  letra(65,68,'N',4,BLUE,YELLOW);
  vga_getch();
  borrarc();
  titulo();
}

void jrandom(void)
{
  int hour,min,sec,day,month,year;
  struct tm *tm_ptr;
  time_t the_time;
  (void) time(&the_time);
  tm_ptr = gmtime(&the_time);
  hour=tm_ptr->tm_hour+1; // (0-23)+1
  min=tm_ptr->tm_min+1; // (0-59)+1
  sec=tm_ptr->tm_sec+1; // (0-61)+1
  day=tm_ptr->tm_mday; // 1-31
  month=tm_ptr->tm_mon+1; // (0-11)+1 
  year=tm_ptr->tm_year; // 1900-
  srandom(min*sec+hour*day+month+year);
}

void inicializar(void)
{
  int x,y;
  for(x=0;x<=12;x++)
	for(y=0;y<=31;y++)
	  m[x][y]=0;
}

int verifica(int a, int b)
{
  int x,y;
  for(x=0;x<=3;x++)
   for(y=0;y<=3;y++)
	 if (m[a+x][b+y]!=0 && f[x][y]==1) return 0; /* No puede bajar */
return 1;
}

int bajar()
{ 
  int x,y,flag;
  for (y=30;y>=1;y--)
	{
	  flag=0;
	  for(x=1;x<=10;x++)
		{
		  if (m[x][y]==0) x=15;
		}
	  if (x==11) { flag=1; break; }
	}
  if (flag==1) return y;
  else return 0;
}

void bajar_m(int w)
{
  int x,y;
  for(x=1;x<=10;x++)
	{
	  cuadro(x,w,GRAY,LIGHTGRAY);
	}
  msleep(350000);  //delay de bajar y borrar linea
  for(y=w;y>=1;y--)
	for(x=1;x<=10;x++)
	  {
		m[x][y]=m[x][y-1];
		if (m[x][y]!=BLACK) cuadro(x,y,m[x][y],WHITE);
		else cuadro(x,y,BLACK,BLACK);
	  }
}

int falling()
{
   int flag,flag1,x,y,w,a=4,b,bb,f,cont;
   char c;
   crear_ficha(rand()%7+1); /* Numero aleatorio entre 1 y 7 */
   level=level+0.02; // Increase the level
   rotar_der();
   for (b=1;b<=30;b++)
    {
       flag=flag1=0; cont=0;
       while(cont <= 13-level)  // Delay level 
         { 
	    if (verifica(a,b)==1) /* La ficha puede ser dibujada en esa posicion */
	      {
		if (cont==0) { borrar(a,b-1); dibujar(a,b); }
		c='0';
		if (kbhit()) c=readch();
		if (c=='p') pausa();  // Pause the game
		if (c=='e')  // End the game
		  {
		     borrar(a,b); // Borra la ficha
		     return 2;
		  }
		if (c=='2')
		  {
		     f=0; bb=b;
		     while (verifica(a,bb)==1) { bb++; f=1; }
		     if (f==1)
		       {
		         borrar(a,b);
			 dibujar(a,bb-1);
			 fijar(a,bb-1);
			 score++; pscore();
			 while (bajar()!=0) { bajar_m(bajar()); score=score+5; pscore(); }
			 flag1=1;
			 b=30;
		       }
		  }
		if (c=='4' && verifica(a-1,b)==1)
		  {
		     borrar(a,b);
		     a--;
		     dibujar(a,b);
		  }
		if (c=='6'&& verifica(a+1,b)==1)
		  {
		     borrar(a,b);
		     a++;
		     dibujar(a,b);
		  }
		 if (c=='5')
		  {
		     rotar_izq();
		     if (verifica(a,b)==1) { borrar(a,b); dibujar(a,b); }
		     else rotar_der();
		  }
		 if (c=='8')
		  {
		     rotar_der();
		     if (verifica(a,b)==1) { borrar(a,b); dibujar(a,b); }
		     else rotar_izq();
		  }
	         if (flag1==1) break;
	      } 
	    else    /* La ficha no puede ser dibujada en esa posicion */
	      {
		 flag=1;
		 if (b==1) return 1; /* Perdio el juego */
		 if (b>1) { fijar(a,b-1); score++; pscore(); }
		 while (bajar()!=0)
		   {
		      score=score+5;
		      pscore();
		      bajar_m(bajar());
		   }
	      }
	    if (flag==1) break;  
	    if (flag1==1) break;
	    msleep(1);  // One microsecond delay
	    cont++;
	 }
       if (flag1==1) break;
       if (flag==1) break;
    }
  return 0;
}

void main()
{
 int w;
 char c='0';
 vga_setmode(G640x480x256);
 jrandom(); // Randomize
 inicializar();
 marco();
 init_keyboard();
 while (c!='x') // Exit
   {
	 c=vga_getch();
	 if (c=='s') // Start the game
	   {
		 w=0;
		 level=0;
		 score=0; pscore();
		 borrargo();
		 inicializar();
		 marco();
		 while (w==0) w=falling();
		 pgameover();
	   }
	 if (c=='i') informacion(); // Info
   }
 vga_setmode(TEXT);
 close_keyboard();
 exit(0);
}
